from __future__ import annotations
from .node import _Node

class _DoublyNode(_Node):
    
    def __init__(self, item, next_node=None, previous_node=None):
        self.__prev = previous_node
        super().__init__(item, next_node)

    def prev(self) -> _DoublyNode:
        return self.__prev

    def set_prev(self, new_node:_DoublyNode) -> _DoublyNode:
        self.__prev = new_node
        return self


class _Negativecycle(Exception):
    '''
    An exception that will be called when a graph exhibits characteristics of negative weight cycles.
    '''
    def __init__(self, message):
        super().__init__(message)


from __future__ import annotations
from jellybeans.exceptions.graph_property import _GraphProperty
from .private_funcs import _is_connected_n_acyclic, _topo_sort


class Graph:
    '''
    Basic Graph data structure.
    '''

    def __init__(self) -> None:
        self.__adj_list = {}
        self.__reachable_by = {}
        self.__unweighted = [True, None]
        self.__positive_weighted = True

    def add_vertex(self, node_id:int) -> None:
        '''
        Add a vertex into the graph 

        Args:
            node_id: Unique identifier of the vertice
        '''
        if node_id not in self.__adj_list and node_id not in self.__reachable_by:
            self.__adj_list[node_id] = []
            self.__reachable_by[node_id] = []
        else:
            raise _GraphProperty("Vertex already present in graph")

    def delete_vertex(self, vertex:int) -> None:
        '''
        Delete a vertex from the graph 

        Args:
            vertex: Unique identifier of the vertice
        '''
        if vertex not in self.__adj_list or vertex not in self.__reachable_by:
            raise _GraphProperty(f"Vertex {vertex} not present in graph")
        reachable = self.can_reach(vertex)
        connected = self.neighbours_of(vertex)

        for tex in reachable:
            self.delete_edge([tex, vertex])

        for node in connected:
            self.delete_edge([vertex, node])

        del self.__adj_list[vertex]
        del self.__reachable_by[vertex]
        
    def list_vertices(self) -> list:
        '''
        List the vertices in the graph 

        Returns:
            A list containing the unique identifiers of the vertices
        '''
        return list(self.__adj_list.keys())

    def num_vertices(self) -> int:
        '''
        Find the number of vertices in the graph 

        Returns:
            The number of vertices present in the graph
        '''
        return len(self.__adj_list)

    def add_edge(self, v_from:int, v_to:int, weight:int = 1) -> None:
        '''
        Add a directed edge into the graph 

        Args:
            v_from: The origin vertice
            v_to: The destination vertice
            weight: The weight/cost of the edge
        '''
        if v_from == v_to:
            raise _GraphProperty("Edges must connect 2 different vertices together")
        if v_from not in self.__adj_list or v_to not in self.__adj_list:
            raise _GraphProperty(f"Unable to add edge as vertex {v_from} or {v_to} not present in graph")
        edges = self.__adj_list[v_from]
        for v,_ in edges:
            if v == v_to:
                raise _GraphProperty(f"Edge {v_from} -> {v_to} already present in graph")
        edges.append((v_to, weight))
        if self.__unweighted[1] is None:
            self.__unweighted[1] = weight
        else:
            if self.__unweighted[0]:
                self.__unweighted[0] = weight == self.__unweighted[1]

        if self.__positive_weighted:
            self.__positive_weighted = weight >= 0
        self.__reachable_by[v_to].append(v_from)
    
    def add_bidirected_edge(self, v_from:int, v_to:int, weight:tuple = (1,1)) -> None:
        '''
        Add a bi-directed edge into the graph 

        Args:
            v_from: The origin vertice
            v_to: The destination vertice
            weight: The weight/cost of the edge in the form of a tuple
        '''
        self.add_edge(v_from, v_to, weight[0])
        self.add_edge(v_to, v_from, weight[1])
    
    def delete_edge(self, edge:list) -> None:
        '''
        Remove a directed edge from the graph 

        Args:
            edge: A list in the form of --> [source vertice, destination vertice]
        '''
        v_from, v_to = edge

        new_adj_list = []
        edge_counter = 0
        for e in self.__adj_list[v_from]:
            if e[0] != v_to:
                new_adj_list.append(e)
                edge_counter += 1
        if edge_counter == len(self.__adj_list[v_from]):
            raise _GraphProperty(f"Edge {edge} not found in the graph")
        
        self.__adj_list[v_from] = new_adj_list
        self.__reachable_by[v_to].remove(v_from)

    def update_edge_weight(self, edge:list, newWeight:int) -> None:
        '''
        Update the weight of an edge 

        Args:
            edge: A list in the form of --> [source vertice, destination vertice]
            newWeight: new weight of the edge
        '''
        vFrom, vTo = edge
        self.delete_edge(edge)
        self.add_edge(vFrom, vTo, newWeight)

    def neighbours_of(self, vertex:int) -> list:
        '''
        FInd the neighbours of a vertice 

        Args:
            vertex: Source vertex
        Returns:
            A list containing the neighbours
        '''
        return [x[0] for x in self.edges_of(vertex)]

    def num_neighbours(self, vertex:int) -> int:
        '''
        FInd the total number of neighbours of a vertice 

        Args:
            vertex: Source vertex
        Returns:
            Sum of neighbours
        '''
        return len(self.neighbours_of(vertex))

    def edges_of(self, vertex:int) -> list:
        '''
        FInd the outgoing edges of a vertex 

        Args:
            vertex: Source vertex
        Returns:
            list containing the outgoing edges
        '''
        return self.__adj_list[vertex].copy()

    def can_reach(self, vertex:int) -> list:
        '''
        FInd the outgoing edges of a vertex 

        Args:
            vertex: Source vertex
        Returns:
            list containing the incoming edges
        '''
        return self.__reachable_by[vertex].copy()

    def to_edge_list(self) -> list:
        '''
        Generate an edge list from the graph 
        
        Returns:
            The edge list
        '''
        res = []
        for vertex in self.list_vertices():
            edges = self.__adj_list[vertex]
            for e in edges:
                complete_edge = (vertex,) + e
                res.append(complete_edge)
        return res

    def transpose(self) -> Graph:
        '''
        Transpose the current graph
        
        Returns:
            The transposed graph
        '''
        new_g = Graph()
        for v in self.list_vertices():
            new_g.add_vertex(v)
        for v_from, v_to, weight in self.to_edge_list():
            new_g.add_edge(v_to, v_from, -weight)
        return new_g
        
    def to_adj_list(self) -> dict:
        '''
        Generate an adjancency list from the graph 
        
        Returns:
            The adjacency list in the form of a dictionary
        '''
        return self.__adj_list.copy()

    def to_adjMatrix(self):
        '''
        Generate an adjacency matrix from the graph 
        
        Returns:
            The adjacency matrix
        '''
        vertices = self.list_vertices()
        mapping = {v:idx for idx,v in enumerate(vertices)}
        res = [[0 for _ in vertices] for _ in vertices]
    
        for v in vertices:
            for neighbor, weight in self.__adj_list[v]:
                res[mapping[v]][mapping[neighbor]] = weight
        return mapping, res

    def is_tree(self) -> bool:
        '''
        Check if the graph is a tree
        
        Returns:
            true if the graph is a tree
        '''
        vertices = self.list_vertices()
        for v in vertices:
            res = _is_connected_n_acyclic(vertices, self.__adj_list, v)
            if not res:
                return False
        return True
        
    def is_unweighted(self) -> bool:
        '''
        Check if the graph is unweighted
        
        Returns:
            true if the graph is unweighted
        '''
        return self.__unweighted[0]

    def is_positive(self) -> bool:
        '''
        Check if the graph contains only positive edge weights
        
        Returns:
            true if the graph has no negative edge weights
        '''
        return self.__positive_weighted

    def is_dag(self) -> bool:
        '''
        Check if the graph is a Directed Acyclic Graph
        
        Returns:
            true if the graph is a directed acyclic graph
        '''
        toposort = _topo_sort(self.list_vertices(), self.to_adj_list(), self.to_edge_list())
        
        if len(toposort) < len(self.__adj_list):
            return False
        idxs = {toposort[idx]: idx for idx in range(len(toposort))}
        for v,edges in self.__adj_list.items():
            for edge,_ in edges:
                idx_v = idxs[v]
                idx_w = idxs[edge]
                if idx_v > idx_w:
                    return False
        return True


'''
TO DEVELOPERS: DO NOT USE THESE FUNCTION WITHOUT UNDERSTANDING ITS FULL FUNCTIONALITY.
'''
from typing import Any, Callable

def swap(lst:list, idx1:int, idx2:int) -> None:
    '''
    Description:
        Private method which swaps two elements in a list
    Args:
        lst: list of elements
        idx1: index of first element to be swapped
        idx2: index of second element to be swapped
    '''
    temp = lst[idx1]
    lst[idx1] = lst[idx2]
    lst[idx2] = temp

def merge(lst:list, start:int, mid:int, end:int, key:Callable) -> None:
    '''
    Description:
        Underlying method two merge 2 lists based on a key
    Args:
        lst: List of elements to be merged
        start: Starting index of the first list to be merged
        mid: Stopping index of the first list to be merged (inclusive)
        end: Stopping index of the second list to be merged
        key: Sorting key
    '''
    temp = []
    left = start
    right = mid+1
    while left <= mid and right <= end:
        if key(lst[left]) <= key(lst[right]):
            temp.append(lst[left])
            left += 1
        else:
            temp.append(lst[right])
            right += 1

    while left <= mid:
        temp.append(lst[left])
        left += 1

    while right <= end:
        temp.append(lst[right])
        right += 1
    for i in range(len(temp)):
        lst[start + i] = temp[i]

def partition(lst:list, start:int, end:int, key:callable) -> int:
    '''
    Description:
        This function seperates elements into 2 different groups.
    Args:
        lst: List of elements to be partitioned
        start: Starting index of elements to partition
        end: Stopping index of elements to partition
        key: Sorting key
    Returns:
        The index of the reference element used to partition the elements
    '''
    pivot = lst[start]
    margin = start
    for idx in range(start+1, end+1):
        if key(lst[idx]) < key(pivot):
            margin += 1
            swap(lst, idx, margin)
    swap(lst, start, margin)
    return margin


from typing import Any
from jellybeans.structures import TailedLinkedList

class Queue:
    '''
    Queue is a FIFO Data structure
    '''
    def __init__(self):
        self.__size = 0
        self.__queue = TailedLinkedList()

    def is_empty(self) -> bool:
        '''
        Returns whether a queue is empty

        Returns:
            True if queue is empty
        '''
        return self.__size == 0

    def peek(self) -> Any:
        '''
        Returns:
            element at the front of the queue
        '''
        if self.is_empty():
            return None
        return self.__queue.get(0)

    def dequeue(self) -> Any:
        '''
        Returns and remove the element at the front of the queue

        Returns:
            element at the front of the queue
        '''
        if self.__size == 0:
            raise IndexError("No item available for removal")
        self.__size -= 1
        val = self.__queue.get(0)
        self.__queue.remove_front()
        return val

    def enqueue(self, item:Any) -> None:
        '''
        Add an item to the back of the queue

        Args:
            item: item to be added
        '''
        self.__size += 1
        self.__queue.add_back(item)

    def __len__(self) -> int:
        '''
        Returns the size of the queue
        '''
        return self.__size

    def __str__(self) -> str:
        '''
        Visual Representation of the queue
        '''
        if self.__size == 0:
            return []
        return " -> ".join(self.__queue.map(lambda x: str(x)).to_list())



from __future__ import annotations
from jellybeans.structures import Queue

def _dfs_tree(visited: list, source: int, adj_list:dict, mapping:dict, parent:int) -> bool:
    '''
    This is a modified version of Depth First Search used to check if a graph is a tree 

    Args:
        visited: List to show if a vertex has been visited
        source: Source Vertex
        adj_List: Graph structure stored in the form of an adjacency list
        mapping: Mapping of vertice number to index number
        parent_lst: List of parents of a node
    '''
    visited[mapping[source]] = 1
    for neighbor, _ in adj_list[source]:
        if visited[mapping[neighbor]] == 0:
            if not _dfs_tree(visited, neighbor, adj_list, mapping, source):
                return False
        elif neighbor != parent:
            return False
    return True

def _is_connected_n_acyclic(vertices:list, adj_list:list, source:int):
    '''
    This function checks if a graph is a connected and acyclic graph.
    
    Args:
        vertices: a list of vertices
        adj_list: Adjacency list
        source: source vertex to begin search
    Returns:
        True if the graph is connected and acyclic
    '''
    visited = [0 for v in vertices]
    mapping = {vertices[idx]:idx for idx in range(len(vertices))}
    if not _dfs_tree(visited, source, adj_list, mapping, source):
        return False
    for v in visited:
        if v != 1:
            return False
    return True

def _topo_sort(vertices:list, adj_list:list, edge_list:list) -> list:
    '''
    Performs a topological sort on the graph. Based on Kahn's Algorithm 

    Args:
        graph: Graph object
    Returns:
        List containing a valid topological ordering
    '''
    in_degree = []
    mapping = {}
    toposort = []
    inv_map = [0 for _ in vertices]
    counter = 0
    q = Queue()
    for v in vertices:
        in_degree.append(0)
        mapping[v] = counter
        inv_map[counter] = v
        counter += 1
    for _, vTo,_ in edge_list:
        in_degree[mapping[vTo]] += 1

    for idx in range(len(in_degree)):
        if in_degree[idx] == 0:
            q.enqueue(inv_map[idx])

    while not q.is_empty():
        tex = q.dequeue()
        toposort.append(tex)
        for v,_ in adj_list[tex]:
            in_degree[mapping[v]] -= 1
            if in_degree[mapping[v]] == 0:
                q.enqueue(v)
    return toposort


from __future__ import annotations
from jellybeans.structures import Queue, Graph
from jellybeans.exceptions.negative_cycle import _Negativecycle


def _initializer(visited: bool, parent: bool, mapping: bool, graph: Graph) -> tuple:
    '''
    Initializer function which helps in producing lists used in graph traversals 

    Args:
        visited: init the visited list?
        visited: init the parent list?
        visited: init the mapping dict?
    Returns:
        A tuple containing the lists
    '''
    visited_lst = [] if visited else None
    parent_lst = [] if parent else None
    mapping_dict = {} if mapping else None
    counter = 0
    for v in graph.list_vertices():
        if visited:
            visited_lst.append(0)
        if parent:
            parent_lst.append(-1)
        if mapping:
            mapping_dict[v] = counter
        counter += 1
    return (visited_lst, parent_lst, mapping_dict)


def _BFS(visited: list, parent: list, mapping: dict, source: int, adj_list: dict) -> None:
    '''
    This is a modified version of Breath First Search 

    Args:
        visited: List to show if a vertex has been visited
        parent: List to show the parent of the current vertex
        mapping: Mapping of vertice number to index number
        source: Source Vertex
        adj_List: Graph structure stored in the form of an adjacency list
    '''
    q = Queue()
    q.enqueue(source)
    visited[mapping[source]] = 1
    while not q.is_empty():
        tex = q.dequeue()
        for v, _ in adj_list[tex]:
            if visited[mapping[v]] == 0:
                visited[mapping[v]] = 1
                parent[mapping[v]] = tex
                q.enqueue(v)


def _dfs(visited: list, source: int, adj_list: dict, mapping: dict) -> None:
    '''
    This is a modified version of Depth First Search 

    Args:
        visited: List to show if a vertex has been visited
        source: Source Vertex
        adj_List: Graph structure stored in the form of an adjacency list
        mapping: Mapping of vertice number to index number
    '''
    visited[mapping[source]] = 1
    for neighbor, _ in adj_list[source]:
        if visited[mapping[neighbor]] == 0:
            _dfs(visited, neighbor, adj_list, mapping)


def _path_construction(parent: list, mapping: dict, source: int, destination: int) -> tuple:
    '''
    Finds a valid path from the source vertex to the destination vertex 

    Args:
        parent: A list of the vertex's predescessor
        mapping: Mapping of vertice number to index nu
        source: Source vertex
        destination: Destination vertex
    Returns:
        A tuple containing the valid path
    '''
    path = []
    end = mapping[destination]
    end_v = destination
    source = mapping[source]
    while end != source:
        path.append(end_v)
        end_v = parent[end]
        end = mapping[parent[end]]
    path.append(end_v)
    path.reverse()
    return tuple(path)


def _dfs_topo(visited: list, toposort_arr: list, source: int, adj_list: dict, mapping: dict) -> None:
    '''
    This is a helper function used for the DFS implementation of the topological sort 

    Args:
        visited: List to show if a vertex has been visited
        toposort_arr: List that contains the toposort ordering
        source: Source Vertex
        adj_List: Graph structure stored in the form of an adjacency list
        mapping: Mapping of vertice number to index number
    '''
    visited[mapping[source]] = 1
    for neighbor, _ in adj_list[source]:
        if visited[mapping[neighbor]] == 0:
            _dfs_topo(visited, toposort_arr, neighbor, adj_list, mapping)
    toposort_arr.append(source)


def _relax(vertex_from: int, vertex_to: int, weight: int, parent: list, cost: list) -> None:
    '''
    This is a internal function used to relax weight costs in SSSP Algorithms
    Args:
        vertex_from: Source vertex
        vertex_to: Neighbouring vertex to source
        weight: edge weight connecting vertex_from to vertex_to
        parent: parent list
        cost: cost list
    '''
    if cost[vertex_to] > cost[vertex_from] + weight:
        cost[vertex_to] = cost[vertex_from] + weight
        parent[vertex_to] = vertex_from


def _dfs_sssp_tree(vertex: int, parent: list, cost: list, adj_list: list, mapping: list) -> None:
    '''
    DFS implementation of doing SSSP on trees
    Args:
        vertex: Source vertex
        parent: Parent list
        cost: Cost list
        adj_list: Adjacency list
        mapping: mapping of vertices to indexes
    '''
    for neighbor, weight in adj_list[vertex]:
        if cost[mapping[neighbor]] == 1000000000:
            _relax(mapping[vertex], mapping[neighbor], weight, parent, cost)
            _dfs_sssp_tree(neighbor, parent, cost, adj_list, mapping)
        else:
            if cost[mapping[vertex]] + weight < cost[mapping[neighbor]]:
                raise _Negativecycle("Edge with negative weight detected!")


def _bfs_sssp_unweighted(vertex: int, parent: list, cost: list, adj_list: list, mapping: list) -> None:
    '''
    BFS implementation of doing SSSP on unweighted graphs
    Args:
        vertex: Source vertex
        parent: Parent list
        cost: Cost list
        adj_list: Adjacency list
        mapping: mapping of vertices to indexes
    '''
    q = Queue()
    q.enqueue(vertex)

    while not q.is_empty():
        curr = q.dequeue()
        for neighbor, weight in adj_list[curr]:
            if cost[mapping[neighbor]] == 1000000000:
                cost[mapping[neighbor]] = cost[mapping[curr]] + weight
                parent[mapping[neighbor]] = mapping[curr]
                q.enqueue(neighbor)
            else:
                if cost[mapping[curr]] + weight < cost[mapping[neighbor]]:
                    raise _Negativecycle("Edge with negative weight detected!")


def _floyd_sp(vertices: list, adj_list: list) -> tuple:
    '''
    Floyd Warshall algorithm to find the shortest path estimate for every pair of vertices
    Args:
        vertices: List of vertices
        adj_list: Adjacency list for every matrix
    '''
    mapping = {v: idx for idx, v in enumerate(vertices)}
    res = [[1000000000 for _ in vertices] for _ in vertices]

    for v in vertices:
        res[v][v] = 0
        for neighbor, weight in adj_list[v]:
            res[mapping[v]][mapping[neighbor]] = weight

    for a in range(len(vertices)):
        for b in range(len(vertices)):
            for c in range(len(vertices)):
                if res[b][a] != 1000000000 and res[a][c] != 1000000000:
                    res[b][c] = min(res[b][c], res[b][a] + res[a][c])
    return res, mapping


def _floyd_reachability(vertices: list, adj_list: list) -> tuple:
    '''
    Floyd Warshall algorithm to find reachability for every pair of vertices
    Args:
        vertices: List of vertices
        adj_list: Adjacency list for every matrix
    '''
    mapping = {v: idx for idx, v in enumerate(vertices)}
    res = [[0 for _ in vertices] for _ in vertices]

    for v in vertices:
        for neighbor, _ in adj_list[v]:
            res[mapping[v]][mapping[neighbor]] = 1

    for a in range(len(vertices)):
        for b in range(len(vertices)):
            for c in range(len(vertices)):
                if res[b][a] != 0 and res[a][c] != 0:
                    res[b][c] = 1 if res[b][c] or (
                        res[b][a] and res[a][c]) else 0
    return res, mapping


def _floyd_detect_cycle(vertices: list, adj_list: list) -> tuple:
    '''
    Floyd Warshall algorithm to detect cycles for every pair of vertices
    Args:
        vertices: List of vertices
        adj_list: Adjacency list for every matrix
    '''
    mapping = {v: idx for idx, v in enumerate(vertices)}
    res = [[1000000000 for _ in vertices] for _ in vertices]

    for v in vertices:
        for neighbor, weight in adj_list[v]:
            res[mapping[v]][mapping[neighbor]] = weight

    for a in range(len(vertices)):
        for b in range(len(vertices)):
            for c in range(len(vertices)):
                if res[b][a] != 1000000000 and res[a][c] != 1000000000:
                    res[b][c] = min(res[b][c], res[b][a] + res[a][c])
    return res, mapping
